# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Adds wrapper classes to translate httpyexpect errors and check against
provided exception specs for all API endpoints
"""

from typing import Dict

import requests
from httpyexpect.client import ExceptionMapping, ResponseTranslator


class ResponseExceptionTranslator:
    """Base class providing behaviour and injection point for spec"""

    def __init__(self, *, spec: Dict[int, object]) -> None:
        self._exception_map = ExceptionMapping(spec)

    def handle(self, response: requests.Response):
        """Translate and raise error, if defined by spec"""
        translator = ResponseTranslator(response, exception_map=self._exception_map)
        translator.raise_for_error()



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Global Config Parameters"""

from ghga_service_chassis_lib.config import config_from_yaml
from pydantic import BaseSettings, Field

from ghga_connector import core


@config_from_yaml(prefix="ghga_connector")
class Config(BaseSettings):
    "Global Config Parameters"

    upload_api: str = Field(
        "https://hd-dev.ghga-dev.de/ucs",
        description="URL to the root of the upload controller API.",
    )
    download_api: str = Field(
        "https://hd-dev.ghga-dev.de/drs3/ga4gh/drs/v1",
        description="URL to the root of the DRS-compatible API used for download.",
    )

    max_retries: int = Field(
        core.MAX_RETRIES, description="Number of times to retry failed API calls."
    )
    max_wait_time: int = Field(
        core.MAX_WAIT_TIME,
        description=(
            "Maximal time in seconds to wait before quitting without a download."
        ),
    )
    part_size: int = Field(
        core.DEFAULT_PART_SIZE, description="The part size to use for download."
    )



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Contains Calls of the Presigned URLs in order to Up- and Download Files
"""

import math
from io import BufferedReader
from typing import Iterator, Sequence

import requests

from ghga_connector.core import exceptions
from ghga_connector.core.constants import TIMEOUT
from ghga_connector.core.session import RequestsSession


def download_content_range(
    *,
    download_url: str,
    start: int,
    end: int,
) -> bytes:
    """Download a specific range of a file's content using a presigned download url."""

    headers = {"Range": f"bytes={start}-{end}"}
    try:
        response = RequestsSession.get(download_url, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=download_url)
        raise exceptions.RequestFailedError(url=download_url) from request_error

    status_code = response.status_code
    # 200, if the full file was returned, 206 else
    if status_code in (200, 206):
        return response.content

    raise exceptions.BadResponseCodeError(url=download_url, response_code=status_code)


def calc_part_ranges(
    *, part_size: int, total_file_size: int, from_part: int = 1
) -> Sequence[tuple[int, int]]:
    """
    Calculate and return the ranges (start, end) of file parts as a list of tuples.

    By default it starts with the first part but you may also start from a specific part
    in the middle of the file using the `from_part` argument. This might be useful to
    resume an interrupted reading process.
    """
    # calc the ranges for the parts that have the full part_size:
    full_part_number = math.floor(total_file_size / part_size)
    part_ranges = [
        (part_size * (part_no - 1), part_size * part_no - 1)
        for part_no in range(from_part, full_part_number + 1)
    ]

    if (total_file_size % part_size) > 0:
        # if the last part is smaller than the part_size, calculate it range separately:
        part_ranges.append((part_size * full_part_number, total_file_size - 1))

    return part_ranges


def download_file_parts(
    *,
    download_url: str,
    part_size: int,
    total_file_size: int,
    from_part: int = 1,
    download_range_func=download_content_range,
    calc_ranges_func=calc_part_ranges,
) -> Iterator[bytes]:
    """
    Returns an iterator to obtain the bytes content of a file in a part by part fashion.

    By default it start with the first part but you may also start from a specific part
    in the middle of the file using the `from_part` argument. This might be useful to
    resume an interrupted reading process.
    """

    part_ranges = calc_ranges_func(
        part_size=part_size, total_file_size=total_file_size, from_part=from_part
    )

    for part_start, part_end in part_ranges:
        yield download_range_func(
            download_url=download_url, start=part_start, end=part_end
        )


def read_file_parts(
    file: BufferedReader, *, part_size: int, from_part: int = 1
) -> Iterator[bytes]:
    """
    Returns an iterator to iterate through file parts of the given size (in bytes).

    By default it start with the first part but you may also start from a specific part
    in the middle of the file using the `from_part` argument. This might be useful to
    resume an interrupted reading process.

    Please note: opening and closing of the file MUST happen outside of this function.
    """

    initial_offset = part_size * (from_part - 1)
    file.seek(initial_offset)

    while True:
        file_part = file.read(part_size)

        if len(file_part) == 0:
            return

        yield file_part


def upload_file_part(*, presigned_url: str, part: bytes) -> None:
    """Upload File"""

    try:
        response = RequestsSession.put(presigned_url, data=part, timeout=TIMEOUT)
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=presigned_url)
        raise exceptions.RequestFailedError(url=presigned_url) from request_error

    status_code = response.status_code
    if status_code == 200:
        return

    raise exceptions.BadResponseCodeError(url=presigned_url, response_code=status_code)



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handling seesion initialization for requests"""

import requests
from requests.adapters import HTTPAdapter, Retry


class RequestsSession:
    """Helper clas to make max_retries user configurable"""

    session: requests.Session

    @classmethod
    def configure(cls, max_retries: int):
        """Configure session with exponential backoff retry"""
        with requests.session() as session:
            # can't be negative - should we log this?
            max_retries = max(0, max_retries)

            retries = Retry(
                total=max_retries,
                backoff_factor=2,
                status_forcelist=[500, 502, 503, 504],
            )
            adapter = HTTPAdapter(max_retries=retries)

            session.mount("http://", adapter=adapter)
            session.mount("https://", adapter=adapter)

            cls.session = session

    @classmethod
    def get(cls, *args, **kwargs):
        """Delegate to session method"""
        return cls.session.get(*args, **kwargs)

    @classmethod
    def patch(cls, *args, **kwargs):
        """Delegate to session method"""
        return cls.session.patch(*args, **kwargs)

    @classmethod
    def post(cls, *args, **kwargs):
        """Delegate to session method"""
        return cls.session.post(*args, **kwargs)

    @classmethod
    def put(cls, *args, **kwargs):
        """Delegate to session method"""
        return cls.session.put(*args, **kwargs)



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test file operations"""

from typing import Optional

import pytest

from ghga_connector.core.file_operations import (
    download_content_range,
    download_file_parts,
)
from tests.fixtures.s3 import S3Fixture, get_big_s3_object, s3_fixture  # noqa: F401


@pytest.mark.parametrize(
    "start, end, file_size",
    [
        (0, 20 * 1024 * 1024 - 1, 20 * 1024 * 1024),  # download full file as one part
        (  # download intermediate part:
            5 * 1024 * 1024,
            10 * 1024 * 1024 - 1,
            20 * 1024 * 1024,
        ),
    ],
)
def test_download_content_range(
    start: int,
    end: int,
    file_size: int,
    s3_fixture: S3Fixture,  # noqa: F811
):
    """Test the `download_content_range` function."""
    # prepare state and the expected result:
    big_object = get_big_s3_object(s3_fixture, object_size=file_size)
    download_url = s3_fixture.storage.get_object_download_url(
        object_id=big_object.object_id, bucket_id=big_object.bucket_id
    )
    expected_bytes = big_object.content[start : end + 1]

    # donwload content range with dedicated function:
    obtained_bytes = download_content_range(
        download_url=download_url, start=start, end=end
    )

    assert expected_bytes == obtained_bytes


@pytest.mark.parametrize(
    "from_part",
    [None, 3],
)
def test_download_file_parts(
    from_part: Optional[int],
    s3_fixture: S3Fixture,  # noqa: F811
):
    """Test the `download_file_parts` function."""
    # prepare state and the expected result:
    part_size = 5 * 1024 * 1024
    big_object = get_big_s3_object(s3_fixture)
    total_file_size = len(big_object.content)
    if from_part is not None:
        offset = (from_part - 1) * part_size
        expected_bytes = big_object.content[offset:total_file_size]
    else:
        expected_bytes = big_object.content

    download_url = s3_fixture.storage.get_object_download_url(
        object_id=big_object.object_id, bucket_id=big_object.bucket_id
    )

    # prepare kwargs:
    kwargs = {
        "download_url": download_url,
        "part_size": part_size,
        "total_file_size": total_file_size,
    }
    if from_part is not None:
        kwargs["from_part"] = from_part

    # donwload file parts with dedicated function:
    obtained_bytes = bytes()
    for part in download_file_parts(**kwargs):
        obtained_bytes += part

    assert expected_bytes == obtained_bytes



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Custom Exceptions."""

from pathlib import Path

import requests
import urllib3.exceptions

from ghga_connector.core.constants import MAX_PART_NUMBER


class DirectoryDoesNotExistError(RuntimeError):
    """Thrown, when the specified directory does not exist."""

    def __init__(self, *, output_dir: Path):
        message = f"The directory {output_dir} does not exist."
        super().__init__(message)


class FileAlreadyExistsError(RuntimeError):
    """Thrown, when the specified file already exists."""

    def __init__(self, *, output_file: str):
        message = f"The file {output_file} does already exist."
        super().__init__(message)


class FileDoesNotExistError(RuntimeError):
    """Thrown, when the specified file does not exist."""

    def __init__(self, *, file_path: Path):
        message = f"The file {file_path} does not exist."
        super().__init__(message)


class PubKeyFileDoesNotExistError(RuntimeError):
    """Thrown, when the specified public key file already exists."""

    def __init__(self, *, pubkey_path: Path):
        message = f"The public key file {pubkey_path} does not exist."
        super().__init__(message)


class ApiNotReachableError(RuntimeError):
    """Thrown, when the api is not reachable."""

    def __init__(self, *, api_url: str):
        message = f"The url {api_url} is currently not reachable."
        super().__init__(message)


class RetryTimeExpectedError(RuntimeError):
    """Thrown, when a request didn't contain a retry time even though it was expected."""

    def __init__(self, *, url: str):
        message = (
            f"No `Retry-After` header in response from server following the url: {url}"
        )
        super().__init__(message)


class RequestFailedError(RuntimeError):
    """Thrown, when a request fails without returning a response code"""

    def __init__(self, *, url: str):
        message = f"The request to {url} failed."
        super().__init__(message)


class NoS3AccessMethodError(RuntimeError):
    """Thrown, when a request returns the desired response code, but no S3 Access
    Method"""

    def __init__(self, *, url: str):
        message = f"The request to {url} did not return an S3 Access Method."
        super().__init__(message)


class FileNotRegisteredError(RuntimeError):
    """Thrown, when a request for a file returns a 404 error."""

    def __init__(self, *, file_id: str):
        message = (
            f"The request for the file {file_id} failed, "
            "because this file id does not exist."
        )
        super().__init__(message)


class UploadNotRegisteredError(RuntimeError):
    """Thrown, when a request for a multipart upload returns a 404 error."""

    def __init__(self, *, upload_id: str):
        message = (
            f"The request for the upload with the id '{upload_id}' failed, "
            "because this upload does not exist."
        )
        super().__init__(message)


class BadResponseCodeError(RuntimeError):
    """Thrown, when a request returns an unexpected response code (e.g. 500)"""

    def __init__(self, *, url: str, response_code: int):
        self.response_code = response_code
        message = f"The request to {url} failed with response code {response_code}"
        super().__init__(message)


class NoUploadPossibleError(RuntimeError):
    """Thrown, when a multipart upload currently can't be started (response code 400)"""

    def __init__(self, *, file_id: str):
        message = (
            "It is not possible to start a multipart upload for file with id"
            + f" '{file_id}', because this download is already pending or has been"
            + " accepted."
        )
        super().__init__(message)


class UserHasNoUploadAccessError(RuntimeError):
    """
    Thrown when a user does not have the credentials to get or change
    details of an ongoing upload with a specific upload id
    (response code 403)
    """

    def __init__(self, *, upload_id: str):
        message = (
            "This user is not registered as data submitter "
            f"for the file corresponding to the upload_id '{upload_id}'."
        )
        super().__init__(message)


class UserHasNoFileAccessError(RuntimeError):
    """
    Thrown when a user does not have the credentials for
    a specific file id (response code 403)
    """

    def __init__(self, *, file_id: str):
        message = (
            "This user is not registered as data submitter "
            f"for the file with the id '{file_id}'."
        )
        super().__init__(message)


class CantChangeUploadStatusError(RuntimeError):
    """
    Thrown when the upload status of a file can't be set to the requested status
    (response code 400)
    """

    def __init__(self, *, upload_id: str, upload_status: str):
        message = f"The upload with id '{upload_id}' can't be set to '{upload_status}'."
        super().__init__(message)


class MaxWaitTimeExceededError(RuntimeError):
    """Thrown, when the specified wait time for getting a download url has been
    exceeded."""

    def __init__(self, *, max_wait_time: int):
        message = f"Exceeded maximum wait time of {max_wait_time} seconds."
        super().__init__(message)


class MaxRetriesReachedError(RuntimeError):
    """Thrown, when the specified number of retries has been exceeded."""

    def __init__(self, *, url: str, reason: str):
        message = f"Exceeded maximum retries for '{url}' due to: {reason}."
        super().__init__(message)


class MaxPartNoExceededError(RuntimeError):
    """
    Thrown requesting a part number larger than the maximally possible number of parts.

    This exception is a bug.
    """

    def __init__(self):
        message = f"No more than ({MAX_PART_NUMBER}) file parts can be up-/downloaded."
        super().__init__(message)


def raise_if_max_retries(request_error: requests.exceptions.RequestException, url: str):
    """
    Check if request exception is caused by hitting max retries and raise accordingly
    """
    if isinstance(request_error, requests.exceptions.ConnectionError):
        if request_error.args and isinstance(
            request_error.args[0], urllib3.exceptions.MaxRetryError
        ):
            max_retry_error = request_error.args[0]
            raise MaxRetriesReachedError(
                url=url, reason=str(max_retry_error.reason)
            ) from max_retry_error



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Config utilies."""

from ghga_connector import core
from ghga_connector.config import Config

DEFAULT_TEST_CONFIG = Config(
    upload_api="http:/example.org/upload",
    download_api="http:/example.org/download",
    max_retries=0,
    max_wait_time=2,
    part_size=core.DEFAULT_PART_SIZE,
)


def get_test_config(**kwargs):
    """Get test config params with the defaults being overwritting by the parameter
    passed as kwargs.
    """

    return DEFAULT_TEST_CONFIG.copy(update=kwargs)



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for API Calls"""

from contextlib import nullcontext
from typing import Optional
from unittest.mock import Mock

import pytest

from ghga_connector.core.api_calls import (
    UploadStatus,
    get_part_upload_urls,
    patch_multipart_upload,
)
from ghga_connector.core.exceptions import (
    CantChangeUploadStatusError,
    MaxPartNoExceededError,
    MaxRetriesReachedError,
    UploadNotRegisteredError,
)
from tests.fixtures.mock_api.testcontainer import MockAPIContainer


@pytest.mark.parametrize(
    "bad_url,upload_id,upload_status,expected_exception",
    [
        (False, "pending", UploadStatus.UPLOADED, None),
        (False, "uploaded", UploadStatus.CANCELLED, None),
        (False, "pending", UploadStatus.CANCELLED, CantChangeUploadStatusError),
        (False, "uploadable", UploadStatus.UPLOADED, CantChangeUploadStatusError),
        (False, "not_uploadable", UploadStatus.UPLOADED, UploadNotRegisteredError),
        (True, "uploaded", UploadStatus.UPLOADED, MaxRetriesReachedError),
    ],
)
def test_patch_multipart_upload(
    bad_url: bool,
    upload_id: str,
    upload_status: UploadStatus,
    expected_exception: type[Optional[Exception]],
):
    """
    Test the patch_multipart_upload function
    """
    with MockAPIContainer() as api:
        api_url = "http://bad_url" if bad_url else api.get_connection_url()

        with pytest.raises(  # type: ignore
            expected_exception
        ) if expected_exception else nullcontext():
            patch_multipart_upload(
                api_url=api_url,
                upload_id=upload_id,
                upload_status=upload_status,
            )


@pytest.mark.parametrize(
    "from_part, end_part, expected_exception",
    [
        (None, 10, None),
        (2, 10, None),
        (9999, 10001, MaxPartNoExceededError),
    ],
)
def test_get_part_upload_urls(
    from_part: Optional[int],
    end_part: int,
    expected_exception: type[Optional[Exception]],
):
    """
    Test the `get_part_upload_urls` generator for iterating through signed part urls
    """
    upload_id = "example-upload"
    api_url = "http://my-api.example"
    from_part_ = 1 if from_part is None else from_part

    # mock the function to get a specific part upload url:
    static_signed_url = "http://my-signed-url.example/97982jsdf7823j"
    get_url_func = Mock(return_value=static_signed_url)

    # create the iterator:
    kwargs = {
        "api_url": api_url,
        "upload_id": upload_id,
        "get_url_func": get_url_func,
    }
    if from_part is not None:
        kwargs["from_part"] = from_part
    part_upload_urls = get_part_upload_urls(**kwargs)  # type: ignore

    with (
        pytest.raises(expected_exception)  # type: ignore
        if expected_exception
        else nullcontext()
    ):
        for idx, signed_url in enumerate(part_upload_urls):
            assert static_signed_url == signed_url

            part_no = idx + from_part_
            get_url_func.assert_called_with(
                api_url=api_url, upload_id=upload_id, part_no=part_no
            )

            if part_no >= end_part:
                break



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Contains abstract message display base class
"""
import enum
from abc import ABC, abstractmethod


class AbstractMessageDisplay(ABC):
    """Simple message display base class"""

    @abstractmethod
    def display(self, message: str):
        """Display a message with standard formatting"""
        ...

    @abstractmethod
    def success(self, message: str):
        """Display a message representing information about a successful operation"""
        ...

    @abstractmethod
    def failure(self, message: str):
        """Display a message representing information about a failed operation"""
        ...


class MessageColors(str, enum.Enum):
    """
    Define commmonly used colors for logging
    For a selection of valid colors see click.termui._ansi_colors:
    https://github.com/pallets/click/blob/c96545f6f4ba0eab99de6ec8b4ceb77c9bdb2528/src/click/termui.py#L30
    """

    DEFAULT: str = "white"
    SUCCESS: str = "green"
    FAILURE: str = "red"



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test data"""

from pathlib import Path
from typing import Dict, List

from ghga_service_chassis_lib.object_storage_dao_testing import ObjectFixture
from ghga_service_chassis_lib.utils import TEST_FILE_PATHS


class FileState:
    """_
    File State class for available files
    """

    def __init__(
        self,
        *,
        file_id: str,
        grouping_label: str,
        file_path: Path,
        populate_storage: bool = True,
    ):
        """
        Initialize file state and create imputed attributes.
        Set populate_storage to true in order to upload them to the localstack storage
        """
        self.file_id = file_id
        self.grouping_label = grouping_label
        self.file_path = file_path
        self.populate_storage = populate_storage

        self.storage_objects: List[ObjectFixture] = []
        if self.populate_storage:
            self.storage_objects.append(
                ObjectFixture(
                    file_path=self.file_path,
                    bucket_id=self.grouping_label,
                    object_id=self.file_id,
                )
            )


FILES: Dict[str, FileState] = {
    "file_uploadable": FileState(
        file_id="uploadable",
        grouping_label="inbox",
        file_path=TEST_FILE_PATHS[0],
        populate_storage=False,
    ),
    "file_not_uploadable": FileState(
        file_id="not-uploadable",
        grouping_label="inbox",
        file_path=TEST_FILE_PATHS[1],
        populate_storage=False,
    ),
    "file_with_bad_path": FileState(
        file_id="bad-path",
        grouping_label="inbox",
        file_path=Path("/bad/path.xyz"),
        populate_storage=False,
    ),
    "file_uploaded": FileState(
        file_id="uploaded",
        grouping_label="inbox",
        file_path=TEST_FILE_PATHS[2],
        populate_storage=True,
    ),
    "file_downloadable": FileState(
        file_id="downloadable",
        grouping_label="outbox",
        file_path=TEST_FILE_PATHS[3],
        populate_storage=True,
    ),
    "file_not_downloadable": FileState(
        file_id="not-downloadable",
        grouping_label="outbox",
        file_path=TEST_FILE_PATHS[1],
        populate_storage=False,
    ),
    "file_retry": FileState(
        file_id="retry",
        grouping_label="outbox",
        file_path=TEST_FILE_PATHS[1],
        populate_storage=False,
    ),
}



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test file operations"""

from typing import Optional

import pytest
from ghga_service_chassis_lib.utils import big_temp_file

from ghga_connector.core.file_operations import read_file_parts


@pytest.mark.parametrize("from_part", (None, 3))
def test_read_file_parts(from_part: Optional[int]):
    """Test reading a full file with the `read_file_parts` function."""
    file_size = 20 * 1024 * 1024
    part_size = 5 * 1024 * 1024

    with big_temp_file(file_size) as file:

        # Get the expected content:
        initial_offset = 0 if from_part is None else part_size * (from_part - 1)
        file.seek(initial_offset)
        expected_content = file.read()
        file.seek(0)

        # read the file in parts:
        obtained_content = bytes()
        file_parts = (
            read_file_parts(file, part_size=part_size)
            if from_part is None
            else read_file_parts(file, part_size=part_size, from_part=from_part)
        )

        for part in file_parts:
            obtained_content += part

        assert expected_content == obtained_content



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A testcontainer for running the mock app in the background."""

from pathlib import Path

import requests
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_container_is_ready

APP_MODULE_PATH = Path(__file__).parent.resolve() / "app.py"


class MockAPIContainer(DockerContainer):
    """
    Test container for FastAPI.

    """

    def __init__(
        self,
        s3_download_url: str = "test://download.url",
        s3_upload_url_1: str = "test://upload.url",
        s3_upload_url_2: str = "test://upload.url",
        s3_download_file_size: int = 146,
        image: str = "ghga/fastapi_essentials:0.73.0",
        port: int = 8000,
    ) -> None:
        """Initialize the Fastapi test container.

        Args:
            image (str, optional):
                The docker image from docker hub. Defaults to "ghga/fastapi_essentials:0.73.0".
            port (int, optional):
                The port to reach the FastAPI. Defaults to 8000.
        """
        super(MockAPIContainer, self).__init__(image=image)

        self._port = port

        self.with_exposed_ports(self._port)
        self.with_env("S3_DOWNLOAD_URL", s3_download_url)
        self.with_env("S3_UPLOAD_URL_1", s3_upload_url_1)
        self.with_env("S3_UPLOAD_URL_2", s3_upload_url_2)
        self.with_env("S3_DOWNLOAD_FIELD_SIZE", s3_download_file_size)
        self.with_volume_mapping(host=str(APP_MODULE_PATH), container="/app.py")
        self.with_command(
            f"python3 -m uvicorn --host 0.0.0.0 --port {self._port} --app-dir / app:app"
        )

    def get_connection_url(self) -> str:
        """Returns an HTTP connection URL to the API root."""
        ip = self.get_container_host_ip()
        port = self.get_exposed_port(self._port)
        return f"http://{ip}:{port}"

    @wait_container_is_ready()
    def readiness_probe(self):
        """Test if the RabbitMQ broker is ready."""
        connection_url = self.get_connection_url()
        request = requests.get(f"{connection_url}/ready", timeout=0.5)

        if request.status_code != 204:
            raise RuntimeError("Mock API server not ready.")

    def start(self):
        """Start the test container."""
        super().start()
        self.readiness_probe()
        return self



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fixtures for testing the storage DAO"""

from dataclasses import dataclass
from typing import List

from ghga_service_chassis_lib.object_storage_dao_testing import ObjectFixture
from ghga_service_chassis_lib.s3_testing import (
    S3Fixture,
    s3_fixture_factory,
    upload_file,
)
from ghga_service_chassis_lib.utils import big_temp_file

from . import state

existing_buckets: List[str] = ["inbox", "outbox"]
existing_objects: List[ObjectFixture] = []

for file in state.FILES.values():
    if file.populate_storage:
        for storage_object in file.storage_objects:
            if storage_object.bucket_id not in existing_buckets:
                existing_buckets.append(storage_object.bucket_id)
            existing_objects.append(storage_object)

s3_fixture = s3_fixture_factory(
    existing_buckets=existing_buckets,
    existing_objects=existing_objects,
)


@dataclass
class BigObjectS3Fixture(S3Fixture):
    """Extends the S3Fixture to include information on a big file stored on storage."""

    big_object: ObjectFixture


def get_big_s3_object(
    s3_fixture: S3Fixture, object_size: int = 20 * 1024 * 1024
) -> ObjectFixture:
    """
    Extends the s3_fixture to also include a big file with the specified `file_size` on
    the provided s3 storage.
    """

    with big_temp_file(object_size) as big_file:

        object_fixture = ObjectFixture(
            file_path=big_file.name,
            bucket_id=s3_fixture.existing_buckets[0],
            object_id="big-downloadable",
        )

        # upload file to s3
        assert not s3_fixture.storage.does_object_exist(
            bucket_id=object_fixture.bucket_id, object_id=object_fixture.object_id
        )
        presigned_url = s3_fixture.storage.get_object_upload_url(
            bucket_id=object_fixture.bucket_id,
            object_id=object_fixture.object_id,
        )
        upload_file(
            presigned_url=presigned_url,
            file_path=big_file.name,
            file_md5=object_fixture.md5,
        )

    return object_fixture



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for the core functions of the cli"""

import pytest

from ghga_connector.core.main import check_url


@pytest.mark.parametrize(
    "api_url,wait_time,expected_response",
    # Google has a higher availability than ghga.de
    [("https://www.google.de/", 1000, True), ("https://bad_url", 1000, False)],
)
def test_check_url(api_url: str, wait_time: int, expected_response: bool):
    """
    Test the check_url function
    """
    response = check_url(api_url, wait_time=wait_time)
    assert response == expected_response



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Runs a small fastapi mock server for testing purposes.
All mocks work correclty with file_id == "1".
The drs3 mock sends back a "wait 1 minute" for file_id == "1m"
All other file_ids will fail
"""

import os
from datetime import datetime, timezone
from enum import Enum
from typing import List, Literal

from fastapi import FastAPI, HTTPException, Request, status
from fastapi.responses import JSONResponse, Response
from pydantic import BaseModel

DEFAULT_PART_SIZE = 16 * 1024 * 1024


class UploadStatus(str, Enum):
    """
    Enum for the possible UploadStatus of a specific upload_id
    """

    ACCEPTED = "accepted"
    CANCELLED = "cancelled"
    FAILED = "failed"
    PENDING = "pending"
    REJECTED = "rejected"
    UPLOADED = "uploaded"


class StatePatch(BaseModel):
    """
    Model containing a state parameter. Needed for the UCS patch: /uploads/... api call
    """

    upload_status: UploadStatus


class StatePost(BaseModel):
    """
    Model containing a state parameter. Needed for the UCS post: /uploads api call
    """

    file_id: str


class PresignedPostURL(BaseModel):
    """
    Model containing an url
    """

    url: str


class Checksum(BaseModel):
    """
    A Checksum as per the DRS OpenApi specs.
    """

    checksum: str
    type: Literal["md5", "sha-256"]


class AccessURL(BaseModel):
    """Describes the URL for accessing the actual bytes of the object as per the
    DRS OpenApi spec."""

    url: str


class AccessMethod(BaseModel):
    """An AccessMethod as per the DRS OpenApi spec."""

    access_url: AccessURL
    type: Literal["s3"] = "s3"  # currently only s3 is supported


class UploadProperties(BaseModel):
    """The Upload Properties returned by the UCS post /uploads endpoint"""

    upload_id: str
    file_id: str
    part_size: int


class FileProperties(BaseModel):
    """The File Properties returned by the UCS get /files/{file_id} endpoint"""

    file_id: str
    file_name: str
    md5_checksum: str
    size: int
    grouping_label: str
    creation_date: datetime
    update_date: datetime
    format: str
    current_upload_id: str


class DrsObjectServe(BaseModel):
    """
    A model containing a DrsObject as per the DRS OpenApi specs.
    This is used to serve metadata on a DrsObject (including the access methods) to the
    user.
    """

    file_id: str  # the file ID
    self_uri: str
    size: int
    created_time: str
    updated_time: str
    checksums: List[Checksum]
    access_methods: List[AccessMethod]


class HttpException(Exception):
    """Testing stand in for httpyexpect HttpException without content validation"""

    def __init__(
        self, *, status_code: int, exception_id: str, description: str, data: dict
    ):
        self.status_code = status_code
        self.exception_id = exception_id
        self.description = description
        self.data = data
        super().__init__(description)


app = FastAPI()


@app.exception_handler(HttpException)
async def httpy_exception_handler(request: Request, exc: HttpException):
    """Transform HttpException data into a proper response object"""
    return JSONResponse(
        status_code=exc.status_code,
        content={
            "exception_id": exc.exception_id,
            "description": exc.description,
            "data": exc.data,
        },
    )


@app.get("/ready", summary="readyness_probe")
async def ready():
    """
    Readyness probe.
    """
    return JSONResponse(None, status_code=status.HTTP_204_NO_CONTENT)


@app.get("/objects/{file_id}", summary="drs3_mock")
async def drs3_objects(file_id: str):

    """
    Mock for the drs3 /objects/{file_id} call
    """

    if file_id == "retry":
        return Response(
            status_code=status.HTTP_202_ACCEPTED, headers={"Retry-After": "10"}
        )

    if file_id == "downloadable" or file_id == "big-downloadable":

        return DrsObjectServe(
            file_id=file_id,
            self_uri=f"drs://localhost:8080//{file_id}",
            size=int(os.environ["S3_DOWNLOAD_FIELD_SIZE"]),
            created_time=datetime.now(timezone.utc).isoformat(),
            updated_time=datetime.now(timezone.utc).isoformat(),
            checksums=[Checksum(checksum="1", type="md5")],
            access_methods=[
                AccessMethod(
                    access_url=AccessURL(url=os.environ["S3_DOWNLOAD_URL"]), type="s3"
                )
            ],
        )

    raise HTTPException(
        status_code=404,
        detail=(f'The DRSObject with the id "{file_id}" does not exist.'),
    )


@app.get("/files/{file_id}", summary="ulc_get_files_mock", status_code=200)
async def ulc_get_files(file_id: str):
    """
    Mock for the ulc GET /files/{file_id} call.
    """

    if file_id == "pending":
        return FileProperties(
            file_id=file_id,
            file_name=file_id,
            md5_checksum="",
            size=0,
            grouping_label="inbox",
            creation_date=datetime.utcnow(),
            update_date=datetime.utcnow(),
            format="",
            current_upload_id="pending",
        )

    raise HttpException(
        status_code=404,
        exception_id="fileNotRegistered",
        description=f'The file with the file_id "{file_id}" does not exist.',
        data={"file_id": file_id},
    )


@app.get("/uploads/{upload_id}", summary="ulc_get_uploads_mock", status_code=200)
async def ulc_get_uploads(upload_id: str):
    """
    Mock for the ulc GET /uploads/{upload_id} call.
    """
    if upload_id == "pending":
        return UploadProperties(
            upload_id="pending",
            file_id="pending",
            part_size=DEFAULT_PART_SIZE,
        )

    raise HttpException(
        status_code=404,
        exception_id="noSuchUpload",
        description=f'The upload with the id "{upload_id}" does not exist.',
        data={"upload_id": upload_id},
    )


@app.post("/uploads", summary="ulc_post_uploads_mock", status_code=200)
async def ulc_post_files_uploads(state: StatePost):
    """
    Mock for the ulc POST /uploads call.
    """

    file_id = state.file_id

    if file_id == "uploadable":
        return UploadProperties(
            upload_id="pending",
            file_id=file_id,
            part_size=DEFAULT_PART_SIZE,
        )
    if file_id == "uploadable-16":
        return UploadProperties(
            upload_id="pending",
            file_id=file_id,
            part_size=16 * 1024 * 1024,
        )

    if file_id == "uploadable-8":
        return UploadProperties(
            upload_id="pending",
            file_id=file_id,
            part_size=8 * 1024 * 1024,
        )
    if file_id == "pending":
        raise HttpException(
            status_code=403,
            exception_id="noFileAccess",
            description=f'Can`t start multipart upload for file with file id "{file_id}".',
            data={"file_id": file_id},
        )

    raise HttpException(
        status_code=400,
        exception_id="fileNotRegistered",
        description=f'The file with the file_id "{file_id}" does not exist.',
        data={"file_id": file_id},
    )


@app.post(
    "/uploads/{upload_id}/parts/{part_no}/signed_posts",
    summary="ulc_post_uploads_parts_files_signed_posts_mock",
    status_code=200,
)
async def ulc_post_uploads_parts_files_signed_posts(upload_id: str, part_no: int):
    """
    Mock for the ulc POST /uploads/{upload_id}/parts/{part_no}/signed_posts call.
    """

    if upload_id == "pending":
        if part_no == 1:
            url = os.environ["S3_UPLOAD_URL_1"]
            return {"presigned_url": url}
        if part_no == 2:
            url = os.environ["S3_UPLOAD_URL_2"]
            return {"presigned_url": url}

    raise HttpException(
        status_code=404,
        exception_id="noSuchUpload",
        description=f'The file with the upload id "{upload_id}" does not exist.',
        data={"upload_id": upload_id},
    )


@app.patch("/uploads/{upload_id}", summary="ulc_patch_uploads_mock", status_code=204)
async def ulc_patch_uploads(upload_id: str, state: StatePatch):

    """
    Mock for the ulc PATCH /uploads/{upload_id} call
    """
    upload_status = state.upload_status

    if upload_id == "uploaded":
        if upload_status == UploadStatus.CANCELLED:
            return JSONResponse(None, status_code=status.HTTP_204_NO_CONTENT)

        raise HttpException(
            status_code=400,
            exception_id="uploadNotPending",
            description=f'The upload with id "{upload_id}" can`t be set to "{upload_status}"',
            data={"upload_id": upload_id, "current_upload_status": upload_id},
        )

    if upload_id == "pending":
        if upload_status == UploadStatus.UPLOADED:
            return JSONResponse(None, status_code=status.HTTP_204_NO_CONTENT)

        raise HttpException(
            status_code=400,
            exception_id="uploadStatusChange",
            description=f'The upload with id "{upload_id}" can`t be set to "{upload_status}"',
            data={"upload_id": upload_id, "target_status": upload_status},
        )

    if upload_id == "uploadable":
        raise HttpException(
            status_code=400,
            exception_id="uploadNotPending",
            description=f'The upload with id "{upload_id}" can`t be set to "{upload_status}"',
            data={"upload_id": upload_id, "current_upload_status": upload_id},
        )

    raise HttpException(
        status_code=404,
        exception_id="noSuchUpload",
        description=f'The upload with id "{upload_id}" does not exist',
        data={"upload_id": upload_id},
    )



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Contains calls to the GHGA storage API
"""

import base64
import json
from enum import Enum
from pathlib import Path
from time import sleep
from typing import Dict, Iterator, Tuple, Union

import crypt4gh.keys
import requests
from requests.structures import CaseInsensitiveDict

from ghga_connector.core import exceptions
from ghga_connector.core.constants import MAX_PART_NUMBER, TIMEOUT
from ghga_connector.core.http_translation import ResponseExceptionTranslator
from ghga_connector.core.message_display import AbstractMessageDisplay
from ghga_connector.core.session import RequestsSession

# Constants for clarity of return values
NO_DOWNLOAD_URL = None
NO_FILE_SIZE = None
NO_RETRY_TIME = None


class UploadStatus(str, Enum):
    """
    Enum for the possible statuses of an upload attempt.
    """

    ACCEPTED = "accepted"
    CANCELLED = "cancelled"
    FAILED = "failed"
    PENDING = "pending"
    REJECTED = "rejected"
    UPLOADED = "uploaded"


def initiate_multipart_upload(
    *,
    api_url: str,
    file_id: str,
    pubkey_path: Path,
) -> Tuple[str, int]:
    """
    Perform a RESTful API call to initiate a multipart upload
    Returns an upload id and a part size
    """

    # build url and headers
    url = f"{api_url}/uploads"
    headers = {"Accept": "application/json", "Content-Type": "application/json"}
    public_key = base64.b64encode(crypt4gh.keys.get_public_key(pubkey_path)).decode()

    post_data = {"file_id": file_id, "submitter_public_key": public_key}
    serialized_data = json.dumps(post_data)

    # Make function call to get upload url
    try:
        response = RequestsSession.post(
            url=url, headers=headers, data=serialized_data, timeout=TIMEOUT
        )
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=url)
        raise exceptions.RequestFailedError(url=url) from request_error

    status_code = response.status_code
    if status_code != 200:
        spec = {
            400: {
                "existingActiveUpload": lambda: exceptions.NoUploadPossibleError(
                    file_id=file_id
                ),
                "fileNotRegistered": lambda: exceptions.FileNotRegisteredError(
                    file_id=file_id
                ),
            },
            403: {
                "noFileAccess": lambda: exceptions.UserHasNoFileAccessError(
                    file_id=file_id
                )
            },
        }
        ResponseExceptionTranslator(spec=spec).handle(response=response)
        raise exceptions.BadResponseCodeError(url=url, response_code=status_code)

    response_body = response.json()

    return response_body["upload_id"], int(response_body["part_size"])


def get_part_upload_url(*, api_url: str, upload_id: str, part_no: int):
    """
    Get a presigned url to upload a specific part
    """

    # build url and headers
    url = f"{api_url}/uploads/{upload_id}/parts/{part_no}/signed_posts"
    headers = {"Accept": "application/json", "Content-Type": "application/json"}

    # Make function call to get upload url
    try:
        response = RequestsSession.post(url=url, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=url)
        raise exceptions.RequestFailedError(url=url) from request_error

    status_code = response.status_code
    if status_code != 200:
        spec = {
            403: {
                "noFileAccess": lambda: exceptions.UserHasNoUploadAccessError(
                    upload_id=upload_id
                )
            },
            404: {
                "noSuchUpload": lambda: exceptions.UploadNotRegisteredError(
                    upload_id=upload_id
                )
            },
        }
        ResponseExceptionTranslator(spec=spec).handle(response=response)
        raise exceptions.BadResponseCodeError(url=url, response_code=status_code)

    response_body = response.json()
    presigned_url = response_body["presigned_url"]

    return presigned_url


def get_part_upload_urls(
    *,
    api_url: str,
    upload_id: str,
    from_part: int = 1,
    get_url_func=get_part_upload_url,
) -> Iterator[str]:
    """
    For a specific mutli-part upload identified by the `upload_id`, it returns an
    iterator to iterate through file parts and obtain the corresponding upload urls.

    By default it start with the first part but you may also start from a specific part
    in the middle of the file using the `from_part` argument. This might be useful to
    resume an interrupted upload process.

    Please note: the upload corresponding to the `upload_id` must have already been
    initiated.

    `get_url_func` only for testing purposes.
    """

    for part_no in range(from_part, MAX_PART_NUMBER + 1):
        yield get_url_func(api_url=api_url, upload_id=upload_id, part_no=part_no)

    raise exceptions.MaxPartNoExceededError()


def patch_multipart_upload(
    *, api_url: str, upload_id: str, upload_status: UploadStatus
) -> None:
    """
    Set the status of a specific upload attempt.
    The API accepts "uploaded" or "accepted",
    if the upload_id is currently set to "pending"
    """

    # build url and headers
    url = f"{api_url}/uploads/{upload_id}"
    headers = {"Accept": "*/*", "Content-Type": "application/json"}
    post_data = {"upload_status": upload_status}
    serialized_data = json.dumps(post_data)

    try:
        response = RequestsSession.patch(
            url=url, headers=headers, data=serialized_data, timeout=TIMEOUT
        )
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=url)
        raise exceptions.RequestFailedError(url=url) from request_error

    status_code = response.status_code
    if status_code != 204:
        spec = {
            400: {
                "uploadNotPending": lambda: exceptions.CantChangeUploadStatusError(
                    upload_id=upload_id, upload_status=upload_status
                ),
                "uploadStatusChange": lambda: exceptions.CantChangeUploadStatusError(
                    upload_id=upload_id, upload_status=upload_status
                ),
            },
            403: {
                "noFileAccess": lambda: exceptions.UserHasNoUploadAccessError(
                    upload_id=upload_id
                )
            },
            404: {
                "noSuchUpload": lambda: exceptions.UploadNotRegisteredError(
                    upload_id=upload_id
                )
            },
        }
        ResponseExceptionTranslator(spec=spec).handle(response=response)
        raise exceptions.BadResponseCodeError(url=url, response_code=status_code)


def get_upload_info(
    *,
    api_url: str,
    upload_id: str,
) -> Dict:
    """
    Get details on a specific upload
    """

    # build url and headers
    url = f"{api_url}/uploads/{upload_id}"
    headers = {"Accept": "*/*", "Content-Type": "application/json"}

    try:
        response = RequestsSession.get(url=url, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=url)
        raise exceptions.RequestFailedError(url=url) from request_error

    status_code = response.status_code
    if status_code != 200:
        spec = {
            403: {
                "noFileAccess": lambda: exceptions.UserHasNoUploadAccessError(
                    upload_id=upload_id
                )
            },
            404: {
                "noSuchUpload": lambda: exceptions.UploadNotRegisteredError(
                    upload_id=upload_id
                )
            },
        }
        ResponseExceptionTranslator(spec=spec).handle(response=response)
        raise exceptions.BadResponseCodeError(url=url, response_code=status_code)

    return response.json()


def get_file_metadata(*, api_url: str, file_id: str) -> Dict:
    """
    Get all file metadata
    """

    # build url and headers
    url = f"{api_url}/files/{file_id}"
    headers = {"Accept": "application/json", "Content-Type": "application/json"}

    try:
        response = RequestsSession.get(url=url, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=url)
        raise exceptions.RequestFailedError(url=url) from request_error

    status_code = response.status_code
    if status_code != 200:
        spec = {
            403: {
                "noFileAccess": lambda: exceptions.UserHasNoFileAccessError(
                    file_id=file_id
                )
            },
            404: {
                "fileNotRegistered": lambda: exceptions.FileNotRegisteredError(
                    file_id=file_id
                )
            },
        }
        ResponseExceptionTranslator(spec=spec).handle(response=response)
        raise exceptions.BadResponseCodeError(url=url, response_code=status_code)

    file_metadata = response.json()

    return file_metadata


def download_api_call(
    *, api_url: str, file_id: str
) -> Union[Tuple[None, None, int], Tuple[str, int, None]]:
    """
    Perform a RESTful API call to retrieve a presigned download URL.
    Returns:
        A tuple of three elements:
            1. the download url
            2. the file size (in bytes)
            3. the retry-time
        If the download url is not available yet, the first two elements are None and
        the retry-time is set.
        Otherwise, only the last element is None while the others are set.
    """

    # build url and headers
    url = f"{api_url}/objects/{file_id}"

    headers = CaseInsensitiveDict(
        {"Accept": "application/json", "Content-Type": "application/json"}
    )

    # Make function call to get upload url
    try:
        response = RequestsSession.get(url=url, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.RequestException as request_error:
        exceptions.raise_if_max_retries(request_error=request_error, url=url)
        raise exceptions.RequestFailedError(url=url) from request_error

    status_code = response.status_code
    if status_code != 200:
        if status_code != 202:
            raise exceptions.BadResponseCodeError(url=url, response_code=status_code)

        headers = response.headers
        if "retry-after" not in headers:
            raise exceptions.RetryTimeExpectedError(url=url)

        return (NO_DOWNLOAD_URL, NO_FILE_SIZE, int(headers["retry-after"]))

    # look for an access method of type s3 in the response:
    response_body = response.json()
    download_url = None
    access_methods = response_body["access_methods"]
    for access_method in access_methods:
        if access_method["type"] == "s3":
            download_url = access_method["access_url"]["url"]
            file_size = response_body["size"]
            break

    if download_url is None:
        raise exceptions.NoS3AccessMethodError(url=url)

    return download_url, file_size, NO_RETRY_TIME


def start_multipart_upload(
    *, api_url: str, file_id: str, pubkey_path: Path
) -> Tuple[str, int]:
    """Try to initiate a multipart upload. If it fails, try to cancel the current upload
    can and then try to initiate a multipart upload again."""

    try:
        multipart_upload = initiate_multipart_upload(
            api_url=api_url,
            file_id=file_id,
            pubkey_path=pubkey_path,
        )
        return multipart_upload
    except exceptions.NoUploadPossibleError as error:
        file_metadata = get_file_metadata(api_url=api_url, file_id=file_id)
        upload_id = file_metadata["current_upload_id"]
        if upload_id is None:
            raise error

        patch_multipart_upload(
            api_url=api_url,
            upload_id=upload_id,
            upload_status=UploadStatus.CANCELLED,
        )

        multipart_upload = initiate_multipart_upload(
            api_url=api_url, file_id=file_id, pubkey_path=pubkey_path
        )

    except Exception as error:
        raise error

    return multipart_upload


def await_download_url(
    *,
    api_url: str,
    file_id: str,
    max_wait_time: int,
    message_display: AbstractMessageDisplay,
) -> Tuple[str, int]:
    """Wait until download URL can be generated.
    Returns a tuple with two elements:
        1. the download url
        2. the file size in bytes
    """

    # get the download_url, wait if needed
    wait_time = 0
    while wait_time < max_wait_time:
        try:
            response_body = download_api_call(api_url=api_url, file_id=file_id)
        except exceptions.BadResponseCodeError as error:
            message_display.failure(
                "The request was invalid and returnd a wrong HTTP status code."
            )
            raise error
        except exceptions.RequestFailedError as error:
            message_display.failure("The request has failed.")
            raise error

        if response_body[0] is not None:
            download_url: str = response_body[0]
            file_size: int = response_body[1]
            return (download_url, file_size)

        retry_time: int = response_body[2]

        wait_time += retry_time
        message_display.display(
            f"File staging, will try to download again in {retry_time} seconds"
        )
        sleep(retry_time)

    raise exceptions.MaxWaitTimeExceededError(max_wait_time=max_wait_time)



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

""" CLI-specific wrappers around core functions."""

from pathlib import Path

import typer

from ghga_connector import core
from ghga_connector.config import Config

config = Config()  # will be patched for testing


class CLIMessageDisplay(core.AbstractMessageDisplay):
    """
    Command line writer message display implementation,
    using different color based on information type
    """

    def display(self, message: str):
        """
        Write message with default color to stdout
        """
        typer.secho(message, fg=core.MessageColors.DEFAULT)

    def success(self, message: str):
        """
        Write message to stdout representing information about a successful operation
        """
        typer.secho(message, fg=core.MessageColors.SUCCESS)

    def failure(self, message: str):
        """
        Write message to stderr representing information about a failed operation
        """
        typer.secho(message, fg=core.MessageColors.FAILURE, err=True)


cli = typer.Typer()


@cli.command()
def upload(  # noqa C901
    *,
    file_id: str = typer.Option(..., help="The id if the file to upload"),
    file_path: Path = typer.Option(..., help="The path to the file to upload"),
    pubkey_path: Path = typer.Argument(
        "./key.pub",
        help="The path to a public key from the key pair that was used to encrypt the "
        + "crypt4gh envelope. Defaults to the file key.pub in the current folder.",
    ),
):
    """
    Command to upload a file
    """
    core.RequestsSession.configure(config.max_retries)

    core.upload(
        api_url=config.upload_api,
        file_id=file_id,
        file_path=file_path,
        message_display=CLIMessageDisplay(),
        pubkey_path=pubkey_path,
    )


@cli.command()
def download(  # pylint: disable=too-many-arguments
    *,
    file_id: str = typer.Option(..., help="The id if the file to upload"),
    output_dir: Path = typer.Option(
        ..., help="The directory to put the downloaded file"
    ),
):
    """
    Command to download a file
    """
    core.RequestsSession.configure(config.max_retries)
    core.download(
        api_url=config.download_api,
        file_id=file_id,
        output_dir=output_dir,
        max_wait_time=config.max_wait_time,
        part_size=config.part_size,
        message_display=CLIMessageDisplay(),
    )



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Main domain logic."""

import os
from pathlib import Path

import requests

from ghga_connector.core import exceptions
from ghga_connector.core.api_calls import (
    UploadStatus,
    await_download_url,
    get_part_upload_urls,
    patch_multipart_upload,
    start_multipart_upload,
)
from ghga_connector.core.constants import MAX_WAIT_TIME
from ghga_connector.core.file_operations import (
    download_file_parts,
    read_file_parts,
    upload_file_part,
)
from ghga_connector.core.message_display import AbstractMessageDisplay


def check_url(api_url, *, wait_time=1000) -> bool:
    """
    Checks, if an url is reachable within a certain time
    """
    try:
        # timeout takes seconds, was ms in curl, convert accordingly
        requests.get(url=api_url, timeout=wait_time / 1000)
    except requests.exceptions.RequestException:
        return False
    return True


def upload(  # noqa C901, pylint: disable=too-many-statements,too-many-branches
    *,
    api_url: str,
    file_id: str,
    file_path: Path,
    message_display: AbstractMessageDisplay,
    pubkey_path: Path,
) -> None:
    """
    Core command to upload a file. Can be called by CLI, GUI, etc.
    """

    if not os.path.isfile(pubkey_path):
        message_display.failure(f"The file {pubkey_path} does not exist.")
        raise exceptions.PubKeyFileDoesNotExistError(pubkey_path=pubkey_path)

    if not os.path.isfile(file_path):
        message_display.failure(f"The file {file_path} does not exist.")
        raise exceptions.FileDoesNotExistError(file_path=file_path)

    if not check_url(api_url):
        message_display.failure(f"The url {api_url} is currently not reachable.")
        raise exceptions.ApiNotReachableError(api_url=api_url)

    try:
        upload_id, part_size = start_multipart_upload(
            api_url=api_url, file_id=file_id, pubkey_path=pubkey_path
        )
    except exceptions.NoUploadPossibleError as error:
        message_display.failure(
            f"This user can't start a multipart upload for the file_id '{file_id}'"
        )
        raise error
    except exceptions.UploadNotRegisteredError as error:
        message_display.failure(
            f"The pending upload for file '{file_id}' does not exist."
        )
        raise error
    except exceptions.UserHasNoUploadAccessError as error:
        message_display.failure(
            f"The user is not registered as a Data Submitter for the file with id '{file_id}'."
        )
        raise error
    except exceptions.FileNotRegisteredError as error:
        message_display.failure(f"The file with the id {file_id} is not registered.")
        raise error
    except exceptions.BadResponseCodeError as error:
        message_display.failure(
            "The request was invalid and returnd a wrong HTTP status code."
        )
        raise error
    except exceptions.CantChangeUploadStatusError as error:
        message_display.failure(f"The file with id '{file_id}' was already uploaded.")
        raise error
    except exceptions.RequestFailedError as error:
        message_display.failure("The request has failed.")
        raise error

    try:
        upload_file_parts(
            api_url=api_url,
            upload_id=upload_id,
            part_size=part_size,
            file_path=file_path,
        )
    except exceptions.MaxRetriesReachedError as error:
        message_display.failure(
            "The upload has failed too many times. The upload was aborted."
        )
        raise error

    try:
        patch_multipart_upload(
            api_url=api_url,
            upload_id=upload_id,
            upload_status=UploadStatus.UPLOADED,
        )
    except exceptions.BadResponseCodeError as error:
        message_display.failure(
            f"The request to confirm the upload with id {upload_id} was invalid."
        )
        raise error
    except exceptions.RequestFailedError as error:
        message_display.failure(f"Confirming the upload with id {upload_id} failed.")
        raise error
    message_display.success(f"File with id '{file_id}' has been successfully uploaded.")


def upload_file_parts(
    *,
    api_url: str,
    upload_id: str,
    part_size: int,
    file_path: Path,
) -> None:
    """
    Uploads a file using a specific upload id via uploading all its parts.
    """

    with open(file_path, "rb") as file:
        file_parts = read_file_parts(file, part_size=part_size)
        upload_urls = get_part_upload_urls(api_url=api_url, upload_id=upload_id)

        for part, upload_url in zip(file_parts, upload_urls):
            upload_file_part(presigned_url=upload_url, part=part)


def download(  # pylint: disable=too-many-arguments
    *,
    api_url: str,
    file_id: str,
    output_dir: Path,
    part_size: int,
    message_display: AbstractMessageDisplay,
    max_wait_time: int = MAX_WAIT_TIME,
) -> None:
    """
    Core command to download a file. Can be called by CLI, GUI, etc.
    """

    if not os.path.isdir(output_dir):
        message_display.failure(f"The directory {output_dir} does not exist.")
        raise exceptions.DirectoryDoesNotExistError(output_dir=output_dir)

    if not check_url(api_url):
        message_display.failure(f"The url {api_url} is currently not reachable.")
        raise exceptions.ApiNotReachableError(api_url=api_url)

    download_url, file_size = await_download_url(
        api_url=api_url,
        file_id=file_id,
        max_wait_time=max_wait_time,
        message_display=message_display,
    )

    # perform the download:

    output_file = os.path.join(output_dir, file_id)
    if os.path.isfile(output_file):
        message_display.failure(f"The file {output_file} already exists.")
        raise exceptions.FileAlreadyExistsError(output_file=output_file)

    try:
        download_parts(
            file_size=file_size,
            download_url=download_url,
            output_file=output_file,
            part_size=part_size,
        )
    except exceptions.MaxRetriesReachedError as error:
        # Remove file, if the download failed.
        os.remove(output_file)
        raise error

    message_display.success(
        f"File with id '{file_id}' has been successfully downloaded."
    )


def download_parts(
    *,
    file_size: int,
    download_url: str,
    output_file: str,
    part_size: int,
) -> None:
    """
    Downloads a file using a specific download_url via uploading all its parts.
    """

    file_parts = download_file_parts(
        download_url=download_url,
        part_size=part_size,
        total_file_size=file_size,
    )
    with open(output_file, "wb") as file:
        for part in file_parts:
            file.write(part)



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for the up- and download functions of the cli"""

import os
import pathlib
from contextlib import nullcontext
from filecmp import cmp
from pathlib import Path
from typing import Optional
from unittest.mock import patch

import pytest
from ghga_service_chassis_lib.utils import big_temp_file

from ghga_connector.cli import download, upload
from ghga_connector.core import exceptions
from ghga_connector.core.constants import DEFAULT_PART_SIZE
from tests.fixtures import state
from tests.fixtures.config import get_test_config
from tests.fixtures.mock_api.testcontainer import MockAPIContainer
from tests.fixtures.s3 import S3Fixture, get_big_s3_object, s3_fixture  # noqa: F401
from tests.fixtures.utils import BASE_DIR

PUBLIC_KEY_FILE = BASE_DIR / "keypair/key.pub"


@pytest.mark.parametrize(
    "file_size,part_size",
    [
        (6 * 1024 * 1024, 5 * 1024 * 1024),
        (12 * 1024 * 1024, 5 * 1024 * 1024),
        (1 * 1024 * 1024, DEFAULT_PART_SIZE),
        (20 * 1024 * 1024, DEFAULT_PART_SIZE),
    ],
)
def test_multipart_download(
    file_size: int,
    part_size: int,
    s3_fixture: S3Fixture,  # noqa F811
    tmp_path: pathlib.Path,
):
    """Test the multipart download of a file"""
    big_object = get_big_s3_object(s3_fixture, object_size=file_size)

    # right now the desired file size is only
    # approximately met by the provided big file:
    file_size_ = len(big_object.content)

    # get s3 download url
    download_url = s3_fixture.storage.get_object_download_url(
        bucket_id=big_object.bucket_id,
        object_id=big_object.object_id,
        expires_after=180,
    )
    with MockAPIContainer(
        s3_download_url=download_url,
        s3_download_file_size=file_size_,
    ) as api:
        api_url = api.get_connection_url()
        with patch(
            "ghga_connector.cli.config",
            get_test_config(download_api=api_url, part_size=part_size),
        ):
            download(
                file_id=big_object.object_id,
                output_dir=tmp_path,
            )
        with open(tmp_path / big_object.object_id, "rb") as file:
            observed_content = file.read()

        assert observed_content == big_object.content


@pytest.mark.parametrize(
    "bad_url,bad_outdir,file_name,expected_exception",
    [
        (True, False, "file_downloadable", exceptions.ApiNotReachableError),
        (False, False, "file_downloadable", None),
        (
            False,
            False,
            "file_not_downloadable",
            exceptions.BadResponseCodeError,
        ),
        (False, False, "file_retry", exceptions.MaxWaitTimeExceededError),
        (False, True, "file_downloadable", exceptions.DirectoryDoesNotExistError),
    ],
)
def test_download(
    bad_url: bool,
    bad_outdir: bool,
    file_name: str,
    expected_exception: type[Optional[Exception]],
    s3_fixture: S3Fixture,  # noqa: F811
    tmp_path: pathlib.Path,
):
    """Test the download of a file"""

    output_dir = Path("/non/existing/path") if bad_outdir else tmp_path

    file = state.FILES[file_name]

    if file.populate_storage:

        download_url = s3_fixture.storage.get_object_download_url(
            bucket_id=file.grouping_label,
            object_id=file.file_id,
            expires_after=60,
        )

    else:
        download_url = ""

    with MockAPIContainer(
        s3_download_url=download_url,
        s3_download_file_size=os.path.getsize(file.file_path),
    ) as api:
        api_url = "http://bad_url" if bad_url else api.get_connection_url()

        with patch(
            "ghga_connector.cli.config",
            get_test_config(download_api=api_url),
        ):
            with pytest.raises(  # type: ignore
                expected_exception
            ) if expected_exception else nullcontext():
                download(file_id=file.file_id, output_dir=output_dir)

        if not expected_exception:
            assert cmp(output_dir / file.file_id, file.file_path)


@pytest.mark.parametrize(
    "bad_url,file_name,expected_exception",
    [
        (True, "file_uploadable", exceptions.ApiNotReachableError),
        (False, "file_uploadable", None),
        (False, "file_not_uploadable", exceptions.FileNotRegisteredError),
        (False, "file_with_bad_path", exceptions.FileDoesNotExistError),
    ],
)
def test_upload(
    bad_url: bool,
    file_name: str,
    expected_exception: type[Optional[Exception]],
    s3_fixture: S3Fixture,  # noqa F811
):
    """Test the upload of a file, expects Abort, if the file was not found"""

    uploadable_file = state.FILES[file_name]

    # initiate upload
    upload_id = s3_fixture.storage.init_multipart_upload(
        bucket_id=uploadable_file.grouping_label,
        object_id=uploadable_file.file_id,
    )

    upload_url = s3_fixture.storage.get_part_upload_url(
        bucket_id=uploadable_file.grouping_label,
        object_id=uploadable_file.file_id,
        upload_id=upload_id,
        part_number=1,
    )

    with MockAPIContainer(s3_upload_url_1=upload_url) as api:
        api_url = "http://bad_url" if bad_url else api.get_connection_url()

        with patch("ghga_connector.cli.config", get_test_config(upload_api=api_url)):
            with pytest.raises(  # type: ignore
                expected_exception
            ) if expected_exception else nullcontext():
                upload(
                    file_id=uploadable_file.file_id,
                    file_path=uploadable_file.file_path.resolve(),
                    pubkey_path=Path(PUBLIC_KEY_FILE),
                )

                s3_fixture.storage.complete_multipart_upload(
                    upload_id=upload_id,
                    bucket_id=uploadable_file.grouping_label,
                    object_id=uploadable_file.file_id,
                )

                assert s3_fixture.storage.does_object_exist(
                    bucket_id=uploadable_file.grouping_label,
                    object_id=uploadable_file.file_id,
                )


@pytest.mark.parametrize(
    "file_size,anticipated_part_size",
    [
        (6 * 1024 * 1024, 8),
        (20 * 1024 * 1024, 16),
    ],
)
def test_multipart_upload(
    file_size: int,
    anticipated_part_size: int,
    s3_fixture: S3Fixture,  # noqa F811
):
    """Test the upload of a file, expects Abort, if the file was not found"""

    bucket_id = s3_fixture.existing_buckets[0]
    file_id = "uploadable-" + str(anticipated_part_size)

    anticipated_part_size = anticipated_part_size * 1024 * 1024

    anticipated_part_quantity = file_size // anticipated_part_size

    if anticipated_part_quantity * anticipated_part_size < file_size:
        anticipated_part_quantity += 1

    # initiate upload
    upload_id = s3_fixture.storage.init_multipart_upload(
        bucket_id=bucket_id,
        object_id=file_id,
    )

    # create presigned url for upload part 1
    upload_url_1 = s3_fixture.storage.get_part_upload_url(
        upload_id=upload_id,
        bucket_id=bucket_id,
        object_id=file_id,
        part_number=1,
    )

    # create presigned url for upload part 2
    upload_url_2 = s3_fixture.storage.get_part_upload_url(
        upload_id=upload_id,
        bucket_id=bucket_id,
        object_id=file_id,
        part_number=2,
    )

    with MockAPIContainer(
        s3_upload_url_1=upload_url_1,
        s3_upload_url_2=upload_url_2,
    ) as api:
        api_url = api.get_connection_url()

        # create big temp file
        with big_temp_file(file_size) as file:
            with patch(
                "ghga_connector.cli.config",
                get_test_config(upload_api=api_url),
            ):
                upload(
                    file_id=file_id,
                    file_path=Path(file.name),
                    pubkey_path=Path(PUBLIC_KEY_FILE),
                )

        # confirm upload
        s3_fixture.storage.complete_multipart_upload(
            upload_id=upload_id,
            bucket_id=bucket_id,
            object_id=file_id,
            anticipated_part_quantity=anticipated_part_quantity,
            anticipated_part_size=anticipated_part_size,
        )
        assert s3_fixture.storage.does_object_exist(
            bucket_id=bucket_id,
            object_id=file_id,
        )
